%{

/*
   expr.l -- PIKT script expression lexer

   $Header: /var/cvs/PIKT/pikt/src/piktd/expr.l,v 1.60 2005/01/04 13:49:04 berto Exp $

   This file is part of PIKT ("Problem Informant/Killer Tool").
   Copyright (C) 1998,1999,2000,2001,2002,2003,2004,2005 by Robert Osterlund.

   PIKT is a trademark of the University of Chicago.  All Rights Reserved.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   PIKT is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
   USA.

   Send problem reports, bouquets and brickbats to:  pikt-bugs@pikt.org
 */

#include "../lib/common.h"
#include "piktd.h"
#include "expr.tab.h"

char _exprlex_[] = "exprlex";

#define RETURN(token) exprlval.sval = mystrdup(yytext, _exprlex_); return token

#undef YY_INPUT
#define YY_INPUT(buf, result, max_size) { \
		signed char c = exprinput(); \
		result = (c == EOF) ? YY_NULL : (buf[0] = c, 1); \
	}

#define EXPERR		if (doeval && ! iserrdef) exprres = ERROR;\
			cp = (signed char *) err

#define EXPNIL		if (doeval && ! iserrdef) exprres = NIL;\
			cp = (signed char *) nil

#define NULSTR		tokcurr = tokcurr->next;\
			cp = (signed char *) NULL;\
			return '\0'

signed char *cp;

char arytypvrndx[32];
signed char vrndx[3];

/* note: inlin, inline & inputline should all be absent from these lists! */

char fnnumnames[] = " abs absval absolutevalue age and ascii average avg ceiling code datevalue day daynumber daysbetween defined endofinput eoi err error even false fclose fileage fileino filelinks fileuid filegid filesize fileatime filemtime filectime find fopen grgid groupgid hour if index inputlength inlen inputnumber innum int integer isalnum isalpha isblank iscntrl isdigit isgraph islower isprint ispunct isspace isupper keys length level max maximum mean median members min minimum minute mode month monthnumber nil not now odd or ord parse pclose popen pid procid processid pprocid pprocessid ppid pwuid passwduid pwgid passwdgid random read round rfind rindex rkeys reversekeys rsearch rule rulenumber search second skeys sortkeys split system timevalue today true trunc truncate val value weekday write year yearday yearweek ";

char fnstrnames[] = " alarm alert ampm basename char checksum chop command dayname dirname dquote err filemode fixed grname groupname grpassword grouppassword grmem groupmem grmembers groupmembers hostname if left level lower ltrim mid middle monthname newline nil proper pwname passwdname pwpassword passwdpassword pwcomment passwdcomment pwgecos passwdgecos pwdir passwddir pwshell passwdshell repeat replace reverse right rtrim script space squote substitute substring substr str string tab task text trim upper ";

void exprerror(const char *);
signed int exprinput(void);

%}

W		[ \t]
D		[0-9]
L		[A-Za-z]
DL		[0-9A-Za-z]

%option noyywrap

%%

"("				{ RETURN(_LP); }
")"				{ RETURN(_RP); }
"["				{ RETURN(_LB); }
"]"				{ RETURN(_RB); }
\"\[\"				{ RETURN(_QLB); }
","				{ RETURN(_CMA); }
"!"				{ RETURN(_NOT); }
"&&"				{ RETURN(_AND); }
"||"				{ RETURN(_OR); }

"+"				{ RETURN(_ADD); }
"-"				{ RETURN(_SUBTRACT); }
"*"				{ RETURN(_MULTIPLY); }
"/"				{ RETURN(_DIVIDE); }
"%"				{ RETURN(_MODULUS); }
"**"				{ RETURN(_EXPONENTIATE); }
"&"				{ RETURN(_BITAND); }
"|"				{ RETURN(_BITOR); }
"^"				{ RETURN(_BITEXOR); }
"~"				{ RETURN(_BITCOMP); }
"<<"				{ RETURN(_BITLSHFT); }
">>"				{ RETURN(_BITRSHFT); }

"."				{ RETURN(_CONCATENATE); }

"=="				{ RETURN(_EQ); }
">"				{ RETURN(_GT); }
"<"				{ RETURN(_LT); }
">="				{ RETURN(_GE); }
"<="				{ RETURN(_LE); }
"!="				{ RETURN(_NE); }
"<>"				{ RETURN(_NE2); }
"!>"				{ RETURN(_NGT); }
"!>="				{ RETURN(_NGE); }
"!<"				{ RETURN(_NLT); }
"!<="				{ RETURN(_NLE); }
"=~"				{ RETURN(_RGXP); }
"=~~"				{ RETURN(_RGXPI); }
"!~"				{ RETURN(_NRGXP); }
"!~~"				{ RETURN(_NRGXPI); }
eq				{ RETURN(_SEQ); }
ne				{ RETURN(_SNE); }
lt				{ RETURN(_SLT); }
gt				{ RETURN(_SGT); }
le				{ RETURN(_SLE); }
ge				{ RETURN(_SGE); }

-e				{ RETURN(_EXIST); }
-z				{ RETURN(_ZEROLEN); }
-f				{ RETURN(_REGFILE); }
-d				{ RETURN(_DIRECTORY); }
-c				{ RETURN(_CHARSPEC); }
-b				{ RETURN(_BLKSPEC); }
-p				{ RETURN(_PIPE); }
-l				{ RETURN(_SYMLINK); }
-S				{ RETURN(_SOCKET); }

#abs|#absval|#absolutevalue	{ RETURN(_ABSVAL); }
#age|#fileage			{ RETURN(_AGE); }
$alarm				{ RETURN(_ALARM); }
$alert				{ RETURN(_ALERT); }
$ampm				{ RETURN(_AMPM); }
#and				{ RETURN(_ANDFN); }
#ascii				{ RETURN(_CODE); }
#avg|#average			{ RETURN(_AVERAGE); }
$basename			{ RETURN(_BASENAME); }
#ceiling			{ RETURN(_CEILING); }
$char				{ RETURN(_CHAR); }
$checksum			{ RETURN(_CHECKSUM); }
$chop				{ RETURN(_CHOP); }
#code				{ RETURN(_CODE); }
$command			{ RETURN(_COMMAND); }
#datevalue			{ RETURN(_DATEVALUE); }
#day				{ RETURN(_DAY); }
$dayname			{ RETURN(_DAYNAME); }
#daynumber			{ RETURN(_DAYNUMBER); }
#daysbetween			{ RETURN(_DAYSBETWEEN); }
#defined			{ iserrdef = TRUE;
				  RETURN(_DEFINED);
				}
$dirname			{ RETURN(_DIRNAME); }
$dquote				{ RETURN(_DQUOTE); }
#eoi|#endofinput		{ RETURN(_ENDOFINPUT); }
#err				{ RETURN(_ERRFNNUM); }
$err				{ RETURN(_ERRFNSTR); }
#error				{ iserrdef = TRUE;
				  RETURN(_ERROR);
				}
#even				{ RETURN(_EVEN); }
#false				{ RETURN(_FALSEFN); }
#fclose				{ RETURN(_FCLOSE); }
$filemode			{ RETURN(_FILEMODE); }
#fileino			{ RETURN(_FILEINO); }
#filelinks			{ RETURN(_FILELINKS); }
#fileuid			{ RETURN(_FILEUID); }
#filegid			{ RETURN(_FILEGID); }
#filesize			{ RETURN(_FILESIZE); }
#fileatime			{ RETURN(_FILEATIME); }
#filemtime			{ RETURN(_FILEMTIME); }
#filectime			{ RETURN(_FILECTIME); }
#find				{ RETURN(_SEARCH); }
$fixed				{ RETURN(_FIXED); }
#fopen				{ RETURN(_MYFOPEN);
				  /* _FOPEN -> _MYFOPEN to avoid a name
				     conflict with a SunOS #define
				   */
				 }
$grname|$groupname		{ RETURN(_GRNAME); }
$grpassword|$grouppassword	{ RETURN(_GRPASSWORD); }
#grgid|#groupgid		{ RETURN(_GRGID); }
$grmem|$grmembers|$groupmem|$groupmembers { RETURN(_GRMEMBERS); }
$hostname			{ RETURN(_HOSTNAME); }
#hour				{ RETURN(_HOUR); }
#if				{ RETURN(_IFFNNUM); }
$if				{ RETURN(_IFFNSTR); }
#index				{ RETURN(_SEARCH); }
$inlin|$inline|$inputline	{ RETURN(_INPUTLINE); }
#inlen|#inputlength		{ RETURN(_INPUTLENGTH); }
#innum|#inputnumber		{ RETURN(_INPUTNUMBER); }
#int|#integer			{ RETURN(_INTEGER); }
#isalnum			{ RETURN(_ISITALNUM); }
#isalpha			{ RETURN(_ISITALPHA); }
#isblank			{ RETURN(_ISITBLANK); }
#iscntrl			{ RETURN(_ISITCNTRL); }
#isdigit			{ RETURN(_ISITDIGIT); }
#isgraph			{ RETURN(_ISITGRAPH); }
#islower			{ RETURN(_ISITLOWER); }
#isprint			{ RETURN(_ISITPRINT); }
#ispunct			{ RETURN(_ISITPUNCT); }
#isspace			{ RETURN(_ISITSPACE); }
#isupper			{ RETURN(_ISITUPPER); }
#keys				{ islistfn = TRUE;
				  vrndxfirst = vrndxcurr = 0;
				  RETURN(_KEYS);
				}
$left				{ RETURN(_LEFT); }
#level				{ RETURN(_LEVELFNNUM); }
$level				{ RETURN(_LEVELFNSTR); }
$ltrim				{ RETURN(_LTRIM); }
#length				{ RETURN(_LENGTH); }
$lower				{ RETURN(_LOWER); }
#max|#maximum			{ RETURN(_MAXIMUM); }
#mean				{ RETURN(_AVERAGE); }
#median				{ RETURN(_MEDIAN); }
#members			{ islistfn = TRUE;
				  vrndxfirst = vrndxcurr = 0;
				  RETURN(_MEMBERS);
				}
$mid|$middle			{ RETURN(_SUBSTRING); }
#min|#minimum			{ RETURN(_MINIMUM); }
#minute				{ RETURN(_MINUTE); }
#mode				{ RETURN(_MODE); }
#month				{ RETURN(_MONTH); }
$monthname			{ RETURN(_MONTHNAME); }
#monthnumber			{ RETURN(_MONTHNUMBER); }
$newline			{ RETURN(_NEWLINE); }
#nil				{ RETURN(_NILFNNUM); }
$nil				{ RETURN(_NILFNSTR); }
#not				{ RETURN(_NOTFN); }
#now				{ RETURN(_NOW); }
#odd				{ RETURN(_ODD); }
#or				{ RETURN(_ORFN); }
#ord				{ RETURN(_CODE); }
#pclose				{ RETURN(_PCLOSE); }
#popen				{ RETURN(_POPEN); }
#procid|#processid|#pid		{ RETURN(_PROCID); }
#pprocid|#pprocessid|#ppid	{ RETURN(_PPROCID); }
$proper				{ RETURN(_PROPER); }
$pwname|$passwdname		{ RETURN(_PWNAME); }
$pwpassword|$passwdpassword	{ RETURN(_PWPASSWORD); }
#pwuid|#passwduid		{ RETURN(_PWUID); }
#pwgid|#passwdgid		{ RETURN(_PWGID); }
$pwgecos|$pwcomment|$passwdgecos|$passwdcomment { RETURN(_PWGECOS); }
$pwdir|$passwddir		{ RETURN(_PWDIR); }
$pwshell|$passwdshell		{ RETURN(_PWSHELL); }
#random				{ RETURN(_RANDOM); }
#read				{ RETURN(_READ); }
$repeat				{ RETURN(_REPEAT); }
$replace			{ RETURN(_REPLACE); }
$reverse			{ RETURN(_REVERSE); }
$right				{ RETURN(_RIGHT); }
#rkeys|#reversekeys		{ islistfn = TRUE;
				  vrndxfirst = vrndxcurr = 0;
				  RETURN(_RKEYS);
				}
#round				{ RETURN(_ROUND); }
#rfind|#rindex|#rsearch		{ RETURN(_RSEARCH); }
$rtrim				{ RETURN(_RTRIM); }
#rule|#rulenumber		{ RETURN(_RULE); }
$script				{ RETURN(_ALARM); }
#search				{ RETURN(_SEARCH); }
#second				{ RETURN(_SECOND); }
#skeys|#sortkeys		{ islistfn = TRUE;
				  vrndxfirst = vrndxcurr = 0;
				  RETURN(_SKEYS);
				}
$space				{ RETURN(_SPACE); }
#split|#parse			{ islistfn = TRUE;
				  vrndxfirst = vrndxcurr = 0;
				  RETURN(_SPLIT);
				}
$squote				{ RETURN(_SQUOTE); }
$substitute			{ RETURN(_SUBSTITUTE); }
$substr|$substring		{ RETURN(_SUBSTRING); }
$str|$string			{ RETURN(_FIXED); }
#system				{ RETURN(_SYSTEM); }
$tab				{ RETURN(_TAB); }
$task				{ RETURN(_TASK); }
$text				{ RETURN(_FIXED); }
#timevalue			{ RETURN(_TIMEVALUE); }
#today				{ RETURN(_TODAY); }
$trim				{ RETURN(_TRIM); }
#true				{ RETURN(_TRUEFN); }
#trunc|#truncate		{ RETURN(_TRUNC); }
$upper				{ RETURN(_UPPER); }
#val|#value			{ RETURN(_VALUE); }
#weekday			{ RETURN(_WEEKDAY); }
#write				{ RETURN(_WRITE); }
#year				{ RETURN(_YEAR); }
#yearday			{ RETURN(_YEARDAY); }
#yearweek			{ RETURN(_YEARWEEK); }

"[[ARYSTR "{D}+"]]"		{ yytext[yyleng-2] = '\0';
				  exprlval.ival = atoi(yytext+9);
				  return _ARYSTR;
				}
"[[ARYNUM "{D}+"]]"		{ yytext[yyleng-2] = '\0';
				  exprlval.ival = atoi(yytext+9);
				  return _ARYNUM;
				}

0+{D}+/[\n\177]			{ exprlval.sval = mystrdup(yytext, _exprlex_);
				  return _STR;
				}

-?{D}+/[\n\177]			{ dpls = 0;
				  exprlval.fval = atof(yytext);
				  return _INT;
				}
-?({D}*\.{D}+)/[\n\177]		{ cp = (signed char *) strchr(yytext,'.')+1;
				  dpls = 0;
				  while (*cp) {
					dpls++;
					cp++;
				  }
				  exprlval.fval = atof(yytext);
				  return _NUM;
				}
-?(({D}+)|({D}*\.{D}+))%	{ dpls = 2;
				  if ((cp = (signed char *) strchr(yytext,'.'))) {
					cp++;
				  	while (*cp) {
						dpls++;
						cp++;
					}
				  }
				  yytext[yyleng-1] = '\0';
				  exprlval.fval = atof(yytext)/100;
				  return _NUM;
				}
-?(({D}+)|({D}*\.{D}+))\/(({D}+)|({D}*\.{D}+)) {
				  cp = (signed char *) strchr(yytext,'/');
				  if (atof(cp+1)) {
					*cp = '\0';
					dpls = 6;	/* arbitrary, i know */
					exprlval.fval = atof(yytext)/atof(cp+1);
					return _NUM;
				  } else {
					logreturn(PIKTLOG, iserrdef ? LOG_WARNING : LOG_ERR, ERRID_DIVIDE_BY_ZERO, __FILE__, __LINE__, _exprlex_, FALSE, "in %s, attempted divide by zero: %s", ALMNAM, yytext);
					return _ERR;
				  }
				}

\000				{ return _NULSTR; }

[^\177\"]+			{ exprlval.sval = mystrdup(yytext, _exprlex_);
				  switch ((signed char) yytext[0]) {
					case ERROR:
						return _ERR;
						break;
					case NIL:
						return _NIL;
						break;
					default:
						return _STR;
						break;
				  }
				}

\"				{ return _QT; }

(\177)+				{ ; }

%%

/* expr error handler
 */
void
exprerror (const char *message)
{

	/* the following kludge gets around the problem of blank lines
	   being passwd to the expr parser in a rule like the following:

	   $line regexp "?*" "$line"

	   when we want to see all (new) lines in a (log) file, for
	   example; need a better solution here
	 */
	if (strcmp(yytext, "?*")) {
		if (isdebug)
			logreturn(PIKTLOG, LOG_DEBUG, ERRID_DEBUG, __FILE__, __LINE__, "exprparse", FALSE, "in %s, expr %s: @%s@%d@", ALMNAM, message, yytext ? yytext : nullstring, yytext ? yytext[0] : NONE);
		else
			logreturn(PIKTLOG, doeval ? LOG_ERR : LOG_WARNING, ERRID_EXPR_PARSE_ERROR, __FILE__, __LINE__, "exprparse", FALSE, "in %s, expr %s: @%s@%d@", ALMNAM, message, yytext ? yytext : nullstring, yytext ? yytext[0] : NONE);
	}
	if (doeval) exprres = ERROR;
	return;
}

/* for the current statement, feed its token list to the input stream;
   if we encounter a special var token, feed the appropriate val
 */
signed int
exprinput (void)
{
	static char fnnam[] = "exprinput";
	static signed char c  = '\0';
	static signed char c1;
	static unsigned char c2;
	static signed char *cp = (signed char *) NULL;
	static unsigned offset = 0 ;
	static int ary = 0;
	int i;

	if (! tokcurr) {
		return EOF;
	}

	/* set next cp */
	if (! cp) {
		offset = 0;
		if ((c = c1 = (signed char) *tokcurr->token)) {
			if (c1 > 0) {
				cp = tokcurr->token;
			} else if (c1 < 0) {
				if (vapndx > MAXVAP-1) {
					logexit(PIKTLOG, doeval ? LOG_ERR : LOG_WARNING, ERRID_ARRAY_INDEX_NESTING_MAX_EXCEEDED, __FILE__, __LINE__, fnnam, FALSE, "in %s, max array index nesting (%d) exceeded", ALMNAM, MAXVAP);
					EXPERR;
				} else {
					c2 = (unsigned char) *((tokcurr->token)+1);
					switch (c1) {
						case VLNOW:
							if ((DATTYP(c2) == FILEHANDLE) ||
							    (DATTYP(c2) == PROCHANDLE)   ) {
								/* no vapndx++ here! */
								valp[vapndx] = valnow;
								sprintf(vrndx, "%d", c2);
								cp = vrndx;
							} else {
								valp[vapndx++] = valnow;
								sprintf(arytypvrndx, "[[%s %d]]", arytyp[DATTYP(c2)], c2);
								cp = arytypvrndx;
							}
							break;
						case VLHST:
							if (HSTERR || ! VALHST) {
								tokcurr = tokcurr->next;
								do {
									if (tokcurr->token[0] == '\[')
										ary++;
									else if (tokcurr->token[0] == ']')
										ary--;
									if (ary)
										tokcurr = tokcurr->next;
								} while (ary);
								if (isverboselog)
									logreturn(PIKTLOG, LOG_WARNING, ERRID_NONEXISTENT_HISTORY_VALUE_REFERENCE, __FILE__, __LINE__, fnnam, FALSE, "in %s, attempt to reference nonexistent history value for var %s%s (%d)", ALMNAM, VARNAM(c2), ARYDIM(c2) ? "[]" : nullstring, c2);
								EXPNIL;
							} else {
								valp[vapndx++] = VALHST;
								sprintf(arytypvrndx, "[[%s %d]]", arytyp[DATTYP(c2)], c2);
								cp = arytypvrndx;
								break;
							}
							break;
						case VLPRC:
							if (innum <= 1) {
								tokcurr = tokcurr->next;
								do {
									if (tokcurr->token[0] == '\[')
										ary++;
									else if (tokcurr->token[0] == ']')
										ary--;
									if (ary)
										tokcurr = tokcurr->next;
								} while (ary);
								if (isverboselog)
									logreturn(PIKTLOG, LOG_WARNING, ERRID_NONEXISTENT_PRECEDING_VALUE_REFERENCE, __FILE__, __LINE__, fnnam, FALSE, "in %s, attempt to reference nonexistent preceding value for var %s%s (%d)", ALMNAM, VARNAM(c2), ARYDIM(c2) ? "[]" : nullstring, c2);
								EXPNIL;
							} else {
								valp[vapndx++] = valprc;
								sprintf(arytypvrndx, "[[%s %d]]", arytyp[DATTYP(c2)], c2);
								cp = arytypvrndx;
							}
							break;
						case VLPRS:
							vapndx++;
							sprintf(arytypvrndx, "[[%s %d]]", arytyp[STRING], c2);
							cp = arytypvrndx;
							break;
						case INLINNOW:
							/* dispose of the [0] */
							for (i=0; i<=2; i++)
								tokcurr = tokcurr->next;
							if (! inlinnow) {
								EXPNIL;
							} else {
								cp = inlinnow;
								if (! (c = *cp)) {
									NULSTR;
								}
							}
							break;
						case INLINPRC:
							/* dispose of the [0] */
							for (i=0; i<=2; i++)
								tokcurr = tokcurr->next;
							if (! inlinprc) {
								EXPNIL;
							} else {
								cp = inlinprc;
								if (! (c = *cp)) {
									NULSTR;
								}
							}
							break;
						case OUTLINNOW:
							/* dispose of the [0] */
							for (i=0; i<=2; i++)
								tokcurr = tokcurr->next;
							if (! outlinnow) {
								EXPNIL;
							} else {
								cp = outlinnow;
								if (! (c = *cp)) {
									NULSTR;
								}
							}
							break;
						case OUTLINPRC:
							/* dispose of the [0] */
							for (i=0; i<=2; i++)
								tokcurr = tokcurr->next;
							if (! outlinprc) {
								EXPNIL;
							} else {
								cp = outlinprc;
								if (! (c = *cp)) {
									NULSTR;
								}
							}
							break;
						case RDLINNOW:
							/* dispose of the [0] */
							for (i=0; i<=2; i++)
								tokcurr = tokcurr->next;
							if (! rdlinnow) {
								EXPNIL;
							} else {
								cp = rdlinnow;
								if (! (c = *cp)) {
									NULSTR;
								}
							}
							break;
						case RDLINPRC:
							/* dispose of the [0] */
							for (i=0; i<=2; i++)
								tokcurr = tokcurr->next;
							if (! rdlinprc) {
								EXPNIL;
							} else {
								cp = rdlinprc;
								if (! (c = *cp)) {
									NULSTR;
								}
							}
							break;
						case SQTPROXY:
							cp = tokcurr->token;
							break;
						default:
							logreturn(PIKTLOG, doeval ? LOG_ERR : LOG_WARNING, ERRID_VAR_OUT_OF_BOUNDS, __FILE__, __LINE__, fnnam, FALSE, "in %s, out-of-bounds var %d", ALMNAM, c1);
							EXPERR;
							break;
					}	/* end of switch (c1) */
				}
			} else /* c1 == 0 */ {	/* should never be reached */
				logreturn(PIKTLOG, doeval ? LOG_ERR : LOG_WARNING, ERRID_VAR_OUT_OF_BOUNDS, __FILE__, __LINE__, fnnam, FALSE, "in %s, out-of-bounds var %d", ALMNAM, 0);
				EXPERR;
			}
		} else /* c1 == 0 */ {		/* null string */
			NULSTR;
		}
	}

	if (cp) {
		if ((c = (signed char) *(cp+offset++)) < 0) {
			if (    (cp == (signed char *)  inlinnow)
			     || (cp == (signed char *)  rdlinnow)
			     || (cp == (signed char *) outlinnow)
			     || (cp == (signed char *)  inlinprc)
			     || (cp == (signed char *)  rdlinprc)
			     || (cp == (signed char *) outlinprc)
			   ) {
				c = CLEANCHAR;
			}
		}
	}
	if (! c) {
		c = '\177';
		tokcurr = tokcurr->next;
		cp = (signed char *) NULL;
	}
	return c;

}
